/*
 * Copyright (C) 2006,2007 Lauri Leukkunen <lle@rahina.org>
 * Portion Copyright (c) 2008 Nokia Corporation.
 * (exec postprocessing code implemented by Lauri T. Aarnio at Nokia)
 *
 * Licensed under LGPL version 2.1, see top level LGPL-2.1 file for details.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include "croco.h"

static int elem_count(char **elems)
{
	int count = 0;
	char **p = elems;
	while (*p) {
		p++; count++;
	}
	return count;
}

/* This must be changed to a configuration file format */
int croco_map_path_for_exec(const char *my_file, char **mapping_result)
{
	char *redir_mapfile=getenv("CROCO_MAPFILE");
	char *redir_bin = NULL;
	FILE *fp;
	char *line = NULL;
	size_t len = 0;
	ssize_t read;

	if (redir_mapfile==NULL) {
		SB_LOG(SB_LOGLEVEL_DEBUG, "croco_map_path_for_exec: no mapping");
		*mapping_result=strdup(my_file);
		return 1;
	}
	fp = fopen(redir_mapfile, "r");
	if (fp == NULL) {
		SB_LOG(SB_LOGLEVEL_WARNING, "cannot access: %s", redir_mapfile );
		*mapping_result=strdup(my_file);
		return 1;
	}
	/* /bin/bash,/croco/bin/bash */
	while ((read = getline(&line, &len, fp)) != -1) {
		char *split=strchr(line, ',' );
		if (split == NULL)
			continue;
		if (strncmp(my_file, line, (split-line)) == 0) {
			char *i;
			SB_LOG(SB_LOGLEVEL_DEBUG, "found: %s", line);
			split++;
			i=split; /* chop away any whitespace */
			while(!isspace(*i)) i++;
			redir_bin = strndup(split,i-split);
			SB_LOG(SB_LOGLEVEL_DEBUG, "redirected to: %s", redir_bin);
			break;
		}
	}
	fclose(fp);
	if (line)
		free(line);

	if (redir_bin !=NULL && access(redir_bin,X_OK) == 0) {
		SB_LOG(SB_LOGLEVEL_INFO, "%s -> %s", my_file, redir_bin );
		*mapping_result=redir_bin;
	} else {
		*mapping_result=strdup(my_file);
	}
	return 0;
}

/* Exec Postprocessing:
*/
int croco_execve_postprocess_native(
	char **mapped_file,
	char **filename,
	const char *binary_name,
	char ***argv,
	char ***envp)
{
	int res=1, new_argc;
	int i=0;
	char *host_libs,*tmp,*ld_so;
	char ld_so_buf[CROCO_MAXPATH];
	char ld_so_basename[CROCO_MAXPATH];
	char **p, **my_argv;;


	if (!argv || !envp) {
		SB_LOG(SB_LOGLEVEL_ERROR,
			"ERROR: croco_argvenvp: (argv || envp) == NULL");
		return -1;
	}
	tmp = getenv("HOST_LD_LIBRARY_PATH");
	if (!tmp) {
		host_libs = strdup(HOST_LIB_PATH);
	} else {
		host_libs = strdup(tmp);
	}

	tmp = getenv("HOST_LD_SO");
	if (!tmp) {
		ld_so = strdup(LDPATH);
	} else {
		ld_so = strdup(tmp);
	}

	memset(ld_so_buf, '\0', CROCO_MAXPATH);
	memset(ld_so_basename, '\0', CROCO_MAXPATH);

	if (readlink(ld_so, ld_so_buf, CROCO_MAXPATH) < 0) {
		if (errno == EINVAL) {
			/* it's not a symbolic link, so use it directly */
			strcpy(ld_so_basename, basename(ld_so));
		} else {
			/* something strange, bail out */
			perror("readlink(ld_so) failed badly. aborting\n");
			return -1;
		}
	} else {
		strcpy(ld_so_basename, basename(ld_so_buf));
	}

	new_argc = elem_count(*argv);
	my_argv = (char **)calloc(new_argc+5,sizeof (char *));
	my_argv[i++] = ld_so;
	my_argv[i++] = "--library-path";
	my_argv[i++] = host_libs;
	my_argv[i++] = *mapped_file;

	for (p=*argv+1; *p; p++)
		my_argv[i++]=*p;

	SB_LOG(SB_LOGLEVEL_DEBUG, "Argv to exec %s %s %s %s ...", my_argv[0], my_argv[1], my_argv[2], my_argv[3]);
	*mapped_file=strdup(ld_so);
	*argv=my_argv;
	return res;
}

int croco_execve_postprocess_cputransp(
	char **mapped_file,
	char **filename,
	const char *binary_name,
	char ***argv,
	char ***envp)
{
	int res=1, new_argc;
	int i=0;
	char *cputran,*tmp;
	char **p, **my_argv;;

	if (!argv || !envp) {
		SB_LOG(SB_LOGLEVEL_ERROR,
			"ERROR: croco_argvenvp: (argv || envp) == NULL");
		return -1;
	}
	/* REVISIT: autodetect arch according to binary */
	tmp = getenv("CROCO_CPUTRANSP");
	if (!tmp) {
		cputran = strdup(CPUTRAN);
	} else {
		cputran = strdup(tmp);
	}
	if (access(cputran,X_OK)!=0) {
		SB_LOG(SB_LOGLEVEL_ERROR,
			"ERROR: cpu transparency file %s not executable", cputran);
		return -1;
	}

	new_argc = elem_count(*argv);
	my_argv = (char **)calloc(new_argc+4,sizeof (char *));
	my_argv[i++] = cputran;
	my_argv[i++] = "-U";
	my_argv[i++] = "LD_PRELOAD"; /*XXX: change to filter out libcroco */
	my_argv[i++] = *mapped_file; /* add argv0 */

	for (p=*argv+1; *p; p++)
		my_argv[i++]=*p;

	SB_LOG(SB_LOGLEVEL_DEBUG, "Transp to exec %s %s %s %s ...", my_argv[0], my_argv[1], my_argv[2], my_argv[3]);
	res=croco_execve_postprocess_native(&cputran,filename,binary_name,&my_argv,envp);
	*mapped_file=strdup(cputran);
	*argv=my_argv;
	return res;
}


/* Map script interpreter:
*/
char *croco_execve_map_script_interpreter(
	const char *interpreter,
	const char *interp_arg,
	const char *mapped_script_filename,
	const char *orig_script_filename,
	char ***argv,
	char ***envp)
{
	char *mapped_interpreter;

	if (!argv || !envp) {
		SB_LOG(SB_LOGLEVEL_ERROR,
			"ERROR: croco_execve_map_script_interpreter: "
			"(argv || envp) == NULL");
		return NULL;
	}

	SB_LOG(SB_LOGLEVEL_NOISE,
		"croco_execve_map_script_interpreter: "
		" interpreter=%s interp_arg=%s "
		"mapped_script_filename=%s orig_script_filename=%s",
		interpreter, interp_arg,
		mapped_script_filename, orig_script_filename);

	mapped_interpreter = NULL;
	char *mapping_result;

	croco_map_path_for_exec(interpreter, &mapping_result);
	mapped_interpreter = strdup(mapping_result);

	return mapped_interpreter;
}
